#!/usr/bin/env python3
"""
This file is part of MitoZ.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

COPYRIGHT © 2019-2022 Guanliang Meng. ALL RIGHTS RESERVED.

"""

import sys
import os
import argparse
import time
import subprocess
from string import Template # Python >=3.6
from pathlib import Path


def add_arguments(parser):
    parser.add_argument('--fq1', metavar='<file>', required=True,
        help="Fastq1 file")

    parser.add_argument('--fq2', metavar='<file>',
        help="Fastq2 file")

    parser.add_argument('--phred64', action='store_true',
        help="Are the fastq phred64 encoded? [%(default)s]")

    parser.add_argument('--outprefix', metavar='<str>', default='out',
        help="output prefix [%(default)s]")

    parser.add_argument('--filter_other_para', metavar='<str>', default='',
        help="other parar. [%(default)s]")

    parser.add_argument('--thread_number', metavar='<int>', type=int, default=4,
        help="thread number [%(default)s]")

    parser.add_argument('--workdir', metavar='<directory>', default='./',
        help="working directory [%(default)s]")

    parser.add_argument('--workdir_done', metavar='<directory>', default='./done',
        help="done directory [%(default)s]")

    parser.add_argument('--workdir_log', metavar='<directory>', default='./log',
        help="log directory [%(default)s]")

    return parser


def filter_rawdata(fq1=None, fq2=None, phred64=False, outprefix='out', other_para='', thread=4, workdir='./', workdir_done='./done/', workdir_log='./log/', logger=None):
    out1 = outprefix + ".clean_R1.fq.gz"
    out2 = outprefix + ".clean_R2.fq.gz"
    report_json = outprefix + ".QC.json"
    report_html = outprefix + ".QC.html"

    filtered_fq1 = os.path.join(workdir, out1)
    filtered_fq2 = os.path.join(workdir, out2)

    if phred64:
        phred64_value = '--phred64'
    else:
        phred64_value = ''

    command_template = Template('''
    # set -vex
    # set -o pipefail
    set -e

    mkdir -p ${workdir_done}
    mkdir -p ${workdir_log}
    mkdir -p ${workdir}
    cd ${workdir}

    task_id="FastpTask"
    done_file="${workdir_done}/$task_id.done"
    log_o_file="${workdir_log}/$task_id.o.log"
    log_e_file="${workdir_log}/$task_id.e.log"

    if [ -f $done_file ]; then
        exit 0
    fi

    if [ -f "${fq1}" ] && [ -f "${fq2}" ];
    then
        fastp ${other_para} \
        -i ${fq1} \
        -I ${fq2} ${phred64} \
        --length_required 50 \
        -o ${out1} \
        -O ${out2} \
        -R "${outprefix}" \
        -j ${report_json} \
        -h ${report_html} \
        -w ${thread} && 1>$log_o_file 2>$log_e_file && touch $done_file
    else
        fastp ${other_para} \
        -i ${fq1} ${phred64} \
        --length_required 50 \
        -o ${out1} \
        -R "${outprefix}" \
        -j ${report_json} \
        -h ${report_html} \
        -w ${thread} && 1>$log_o_file 2>$log_e_file && touch $done_file
    fi
    ''')

    command = command_template.safe_substitute(
        workdir=workdir,
        workdir_done=workdir_done,
        workdir_log=workdir_log,
        fq1=fq1,
        fq2=fq2,
        out1=out1,
        out2=out2,
        other_para=other_para,
        outprefix=outprefix,
        report_json=report_json,
        report_html=report_html,
        thread=thread,
        phred64=phred64_value
        )

    # logger.info("Starting to filter raw data {}".format(command))
    runcmd(command, logger=logger)

    if fq1 and not fq2:
        return os.path.abspath(filtered_fq1), ''
    else:
        return os.path.abspath(filtered_fq1), os.path.abspath(filtered_fq2)


def runcmd(command, logger=None):
    logger.info(command)
    subprocess.run(command, shell=True)

    '''
    try:
        current_time = time.strftime("%Y-%m-%d %H:%M:%S",
                        time.localtime(time.time()))
        print(current_time, "\n", command, "\n", sep="", flush=True)
        subprocess.check_call(command, shell=True)
    except:
        sys.exit("Error occured when running command:\n%s" % command)
    '''


def main(args):
    logger = args.logger
    logger.info("filter.main() get args: {}".format(args))

    if args.fq1 and Path(args.fq1).is_file():
        args.fq1 = os.path.abspath(args.fq1)
    else:
        args.fq1 = ''
    if args.fq2 and Path(args.fq2).is_file():
        args.fq2 = os.path.abspath(args.fq2)
    else:
        args.fq2 = ''

    logger.info("filter.main() get updated args: {}".format(args))

    filtered_fq1, filtered_fq2 = filter_rawdata(
        fq1=args.fq1,
        fq2=args.fq2,
        phred64=args.phred64,
        outprefix=args.outprefix,
        other_para=args.filter_other_para,
        thread=args.thread_number,
        workdir=args.workdir,
        workdir_done=args.workdir_done,
        workdir_log=args.workdir_log,
        logger=logger)

    logger.info('filter_rawdata() return:\nfiltered_fq1: {0}\nfiltered_fq2: {1}'.format(filtered_fq1, filtered_fq2))
    return filtered_fq1, filtered_fq2

